#include "InterpolRestrictMatrix.h"
#include "FiniteElemSpace.h"
#include "DOFVector.h"
#include "Traverse.h"
#include "ElInfo.h"
#include "Mesh.h"
#include "BasisFunction.h"
#include "DOFAdmin.h"
#include "Element.h"
#include "SparseVector.h"
#include "ElementMatrix.h"
#include "DOFMatrix.h"
#include <set>

namespace AMDiS {

  InterpolRestrictMatrix::InterpolRestrictMatrix(int numFcts,
						 int numRows,
						 int numCols,
						 bool type0Used)
    : numFcts_(numFcts),
      numRows_(numRows),
      numCols_(numCols),
      type0_used_(type0Used),
      transposed_(false),
      initialized_(false)
  {
    parentDOFs_ = GET_MEMORY(DegreeOfFreedom, numFcts_);
    child0DOFs_ = GET_MEMORY(DegreeOfFreedom, numFcts_);
    child1DOFs_ = GET_MEMORY(DegreeOfFreedom, numFcts_);

    rowDOFPtrs_ = GET_MEMORY(DegreeOfFreedom*, numRows_);
    colDOFPtrs_ = GET_MEMORY(DegreeOfFreedom*, numCols_);

    type0_rowDOFPtrs_ = 
      type0_used_ ? GET_MEMORY(DegreeOfFreedom*, numRows_) : NULL;
    type0_colDOFPtrs_ = 
      type0_used_ ? GET_MEMORY(DegreeOfFreedom*, numCols_) : NULL;
  }


  InterpolRestrictMatrix::InterpolRestrictMatrix(InterpolRestrictMatrix& matrix,
						 bool transposed)
    : numFcts_(matrix.numFcts_),
      numRows_(matrix.numCols_),
      numCols_(matrix.numRows_),
      parentDOFs_(matrix.parentDOFs_),
      child0DOFs_(matrix.child0DOFs_),
      child1DOFs_(matrix.child1DOFs_),
      rowDOFPtrs_(matrix.colDOFPtrs_),
      colDOFPtrs_(matrix.rowDOFPtrs_),
      type0_used_(matrix.type0_used_),
      type0_rowDOFPtrs_(matrix.type0_colDOFPtrs_),
      type0_colDOFPtrs_(matrix.type0_rowDOFPtrs_),
      transposed_(true),
      initialized_(true)
  {
    FUNCNAME("InterpolRestrictMatrix : copy constructor");
    TEST_EXIT(transposed == true)("not yet for transposed == false\n");

    if(!matrix.initialized_) {
      matrix.initialized_ = true;
      matrix.fillMemory();
    }

    // copy directCopyDOFs_
    std::vector<DegreeOfFreedom>::const_iterator copyIt;
    std::vector<DegreeOfFreedom>::const_iterator copyBegin = 
      matrix.directCopyDOFs_.begin();
    std::vector<DegreeOfFreedom>::const_iterator copyEnd =
      matrix.directCopyDOFs_.end();
    for(copyIt = copyBegin; copyIt != copyEnd; ++copyIt) {
      directCopyDOFs_.push_back(*copyIt);
    }

    // fill matrices
    std::map<int, std::map<int, double> >::const_iterator rowsIt;
    std::map<int, std::map<int, double> >::const_iterator rowsBegin;
    std::map<int, std::map<int, double> >::const_iterator rowsEnd;
    std::map<int, double>::const_iterator colIt;
    std::map<int, double>::const_iterator colBegin;
    std::map<int, double>::const_iterator colEnd;

    // unique row matrix
    rowsBegin = matrix.uniqueColMatrix_.begin();
    rowsEnd = matrix.uniqueColMatrix_.end();
    for(rowsIt = rowsBegin; rowsIt != rowsEnd; ++rowsIt) {
      int row = rowsIt->first;
      colBegin = rowsIt->second.begin();
      colEnd = rowsIt->second.end();
      for(colIt = colBegin; colIt != colEnd; ++colIt) {
	int col = colIt->first;
	double entry = colIt->second;
	uniqueRowMatrix_[col][row] = entry;
      }
    }

    // unique col matrix
    rowsBegin = matrix.uniqueRowMatrix_.begin();
    rowsEnd = matrix.uniqueRowMatrix_.end();
    for(rowsIt = rowsBegin; rowsIt != rowsEnd; ++rowsIt) {
      int row = rowsIt->first;
      colBegin = rowsIt->second.begin();
      colEnd = rowsIt->second.end();
      for(colIt = colBegin; colIt != colEnd; ++colIt) {
	int col = colIt->first;
	double entry = colIt->second;
	uniqueColMatrix_[col][row] = entry;
      }
    }

    // normal matrix
    rowsBegin = matrix.matrix_.begin();
    rowsEnd = matrix.matrix_.end();
    for(rowsIt = rowsBegin; rowsIt != rowsEnd; ++rowsIt) {
      int row = rowsIt->first;
      colBegin = rowsIt->second.begin();
      colEnd = rowsIt->second.end();
      for(colIt = colBegin; colIt != colEnd; ++colIt) {
	int col = colIt->first;
	double entry = colIt->second;
	matrix_[col][row] = entry;
      }
    }
  }


  InterpolRestrictMatrix::~InterpolRestrictMatrix()
  {
    if(!transposed_) {
      FREE_MEMORY(parentDOFs_, DegreeOfFreedom, numFcts_);
      FREE_MEMORY(child0DOFs_, DegreeOfFreedom, numFcts_);
      FREE_MEMORY(child1DOFs_, DegreeOfFreedom, numFcts_);

      FREE_MEMORY(rowDOFPtrs_, DegreeOfFreedom*, numRows_);
      FREE_MEMORY(colDOFPtrs_, DegreeOfFreedom*, numCols_);

      if(type0_used_) {
	FREE_MEMORY(type0_rowDOFPtrs_, DegreeOfFreedom*, numRows_);
	FREE_MEMORY(type0_colDOFPtrs_, DegreeOfFreedom*, numCols_);
      }
    }
  }

  void InterpolRestrictMatrix::mv(DOFVectorBase<double> *src, 
				  DOFVectorBase<double> *dest,
				  int coarseLevel)
  {
    FUNCNAME("InterpolRestrictMatrix::mv()");
    
    if(!initialized_) {
      initialized_ = true;
      fillMemory();
    }

    TEST_EXIT_DBG(src)("src not set\n");
    TEST_EXIT_DBG(dest)("dest not set\n");
    TEST_EXIT_DBG(src->getFESpace() == dest->getFESpace())("invalid feSpaces\n");

    const FiniteElemSpace *feSpace = src->getFESpace();

    TEST_EXIT_DBG(feSpace->getBasisFcts()->getNumber() == numFcts_)
      ("invalid number of basis functions\n");

    // do matrix vector multiplication
    DOFVector<int> visitedRowDOFs(feSpace, "visited row DOFs");
    DOFVector<int> visitedColDOFs(feSpace, "visited col DOFs");    

    const BasisFunction *basFcts = feSpace->getBasisFcts();
    DOFAdmin *admin = feSpace->getAdmin();

    std::map<int, std::map<int, double> >::iterator rowsIt;
    std::map<int, std::map<int, double> >::iterator rowsBegin;
    std::map<int, std::map<int, double> >::iterator rowsEnd;
    std::map<int, double>::iterator colIt;
    std::map<int, double>::iterator colBegin;
    std::map<int, double>::iterator colEnd;
    std::vector<DegreeOfFreedom>::iterator dofIt;
    std::vector<DegreeOfFreedom>::iterator dofBegin = directCopyDOFs_.begin();
    std::vector<DegreeOfFreedom>::iterator dofEnd = directCopyDOFs_.end();

    // traverse mesh
    TraverseStack stack;
    ElInfo *elInfo = stack.traverseFirst(feSpace->getMesh(), -1, 
					 Mesh::CALL_EVERY_EL_PREORDER);
    while(elInfo) {
      Element *element = elInfo->getElement();
      if(elInfo->getLevel() == coarseLevel && !element->isLeaf()) {

	DegreeOfFreedom **rowDOFPtrs;
	DegreeOfFreedom **colDOFPtrs;

	if(type0_used_ && dynamic_cast<ElInfo3d*>(elInfo)->getType() == 0) {
	  rowDOFPtrs = type0_rowDOFPtrs_;
	  colDOFPtrs = type0_colDOFPtrs_;
	} else {
	  rowDOFPtrs = rowDOFPtrs_;
	  colDOFPtrs = colDOFPtrs_;
	}

	Element *child0 = element->getChild(0);
	Element *child1 = element->getChild(1);

	basFcts->getLocalIndices(element, admin, parentDOFs_);
	basFcts->getLocalIndices(child0, admin, child0DOFs_);
	basFcts->getLocalIndices(child1, admin, child1DOFs_);

	// direct copy dofs from src to dest if necessary
	if(src != dest) {
	  for(dofIt = dofBegin; dofIt != dofEnd; ++dofIt) {
	    DegreeOfFreedom rowDOF = *(rowDOFPtrs[*dofIt]);
	    if(visitedRowDOFs[rowDOF] == 0) {
	      DegreeOfFreedom colDOF = *(colDOFPtrs[*dofIt]);
	      (*dest)[rowDOF] = (*src)[colDOF];
	      visitedRowDOFs[rowDOF] = 1;
	    }
	  }
	}

	// treat unique rows
	rowsBegin = uniqueRowMatrix_.begin();
	rowsEnd = uniqueRowMatrix_.end();
	for(rowsIt = rowsBegin; rowsIt != rowsEnd; ++rowsIt) {
	  int row = rowsIt->first;
	  DegreeOfFreedom rowDOF = *(rowDOFPtrs[row]);
	  if(visitedRowDOFs[rowDOF] == 0) {
	    visitedRowDOFs[rowDOF] = 1;
	    (*dest)[rowDOF] = 0.0;
	    colBegin = rowsIt->second.begin();
	    colEnd = rowsIt->second.end();
	    for(colIt = colBegin; colIt != colEnd; ++colIt) {
	      int col = colIt->first;
	      DegreeOfFreedom colDOF = *(colDOFPtrs[col]);
	      double entry = colIt->second;
	      (*dest)[rowDOF] += entry * (*src)[colDOF];
	    }
	  }
	}

	// treat unique cols
	std::set<DegreeOfFreedom> elementColDOFs;
	rowsBegin = uniqueColMatrix_.begin();
	rowsEnd = uniqueColMatrix_.end();
	for(rowsIt = rowsBegin; rowsIt != rowsEnd; ++rowsIt) {
	  int row = rowsIt->first;
	  DegreeOfFreedom rowDOF = *(rowDOFPtrs[row]);
	  colBegin = rowsIt->second.begin();
	  colEnd = rowsIt->second.end();
	  for(colIt = colBegin; colIt != colEnd; ++colIt) {
	    int col = colIt->first;
	    DegreeOfFreedom colDOF = *(colDOFPtrs[col]);
	    if(visitedColDOFs[colDOF] == 0) {
	      elementColDOFs.insert(colDOF);
	      double entry = colIt->second;
	      (*dest)[rowDOF] += entry * (*src)[colDOF];
	    }
	  }
	}

	std::set<DegreeOfFreedom>::iterator setIt;
	std::set<DegreeOfFreedom>::iterator setBegin = elementColDOFs.begin();
	std::set<DegreeOfFreedom>::iterator setEnd = elementColDOFs.end();
	for(setIt = setBegin; setIt != setEnd; ++setIt) {
	  visitedColDOFs[*setIt] = 1;
	}

	// treat other dofs
	rowsBegin = matrix_.begin();
	rowsEnd = matrix_.end();
	for(rowsIt = rowsBegin; rowsIt != rowsEnd; ++rowsIt) {
	  int row = rowsIt->first;
	  DegreeOfFreedom rowDOF = *(rowDOFPtrs[row]);
	  colBegin = rowsIt->second.begin();
	  colEnd = rowsIt->second.end();
	  for(colIt = colBegin; colIt != colEnd; ++colIt) {
	    int col = colIt->first;
	    DegreeOfFreedom colDOF = *(colDOFPtrs[col]);
	    double entry = colIt->second;
	    (*dest)[rowDOF] += entry * (*src)[colDOF];
	  }
	}
      }
      elInfo = stack.traverseNext(elInfo);
    }
  }


  void InterpolRestrictMatrix::assembleDOFMatrix(DOFMatrix *dest,
						 int coarseLevel)
  {
    FUNCNAME("InterpolRestrictMatrix::assembleDOFMatrix()");
    TEST_EXIT_DBG(dest)("dest not set\n");
    
    if(!initialized_) {
      initialized_ = true;
      fillMemory();
    }

    dest->clear();

    const FiniteElemSpace *feSpace = dest->getRowFESpace();

    TEST_EXIT_DBG(feSpace->getBasisFcts()->getNumber() == numFcts_)
      ("invalid number of basis functions\n");

    DOFVector<int> visitedRowDOFs(feSpace, "visited row DOFs");
    DOFVector<int> visitedColDOFs(feSpace, "visited col DOFs");    

    const BasisFunction *basFcts = feSpace->getBasisFcts();
    DOFAdmin *admin = feSpace->getAdmin();

    std::map<int, std::map<int, double> >::iterator rowsIt;
    std::map<int, std::map<int, double> >::iterator rowsBegin;
    std::map<int, std::map<int, double> >::iterator rowsEnd;
    std::map<int, double>::iterator colIt;
    std::map<int, double>::iterator colBegin;
    std::map<int, double>::iterator colEnd;
    std::vector<DegreeOfFreedom>::iterator dofIt;
    std::vector<DegreeOfFreedom>::iterator dofBegin = directCopyDOFs_.begin();
    std::vector<DegreeOfFreedom>::iterator dofEnd = directCopyDOFs_.end();

    ElementMatrix elementMatrix(numRows_, numCols_);

    int i;
    TraverseStack stack;
    ElInfo *elInfo = stack.traverseFirst(feSpace->getMesh(), -1, 
					 Mesh::CALL_EVERY_EL_PREORDER);

    while (elInfo) {
      Element *element = elInfo->getElement();
      if (!element->isLeaf()) {
	if (elInfo->getLevel() == coarseLevel) {

	  DegreeOfFreedom **rowDOFPtrs;  
	  DegreeOfFreedom **colDOFPtrs;  

	  if (type0_used_ && dynamic_cast<ElInfo3d*>(elInfo)->getType() == 0) {
	    rowDOFPtrs = type0_rowDOFPtrs_;
	    colDOFPtrs = type0_colDOFPtrs_;
	  } else {
	    rowDOFPtrs = rowDOFPtrs_;
	    colDOFPtrs = colDOFPtrs_;
	  }

	  Element *child0 = element->getChild(0);
	  Element *child1 = element->getChild(1);

	  basFcts->getLocalIndices(element, admin, parentDOFs_);
	  basFcts->getLocalIndices(child0, admin, child0DOFs_);
	  basFcts->getLocalIndices(child1, admin, child1DOFs_);

	  // init element matrix
	  elementMatrix.set(0.0);
	  elementMatrix.rowIndices.resize(numRows_);
	  for (i = 0; i < numRows_; i++) {
	    elementMatrix.rowIndices[i] = *(rowDOFPtrs[i]);
	  }
	  elementMatrix.colIndices.resize(numCols_);
	  for (i = 0; i < numCols_; i++) {
	    elementMatrix.colIndices[i] = *(colDOFPtrs[i]);
	  }

	  // treat direct-copy dofs
	  for (dofIt = dofBegin; dofIt != dofEnd; ++dofIt) {
	    DegreeOfFreedom rowDOF = *(rowDOFPtrs[*dofIt]);
	    if (visitedRowDOFs[rowDOF] == 0) {
	      elementMatrix[*dofIt][*dofIt] = 1.0;
	      visitedRowDOFs[rowDOF] = 1;
	    }
	  }

	  // treat unique rows
	  rowsBegin = uniqueRowMatrix_.begin();
	  rowsEnd = uniqueRowMatrix_.end();
	  for(rowsIt = rowsBegin; rowsIt != rowsEnd; ++rowsIt) {
	    int row = rowsIt->first;
	    DegreeOfFreedom rowDOF = *(rowDOFPtrs[row]);
	    if(visitedRowDOFs[rowDOF] == 0) {
	      visitedRowDOFs[rowDOF] = 1;
	      colBegin = rowsIt->second.begin();
	      colEnd = rowsIt->second.end();
	      for(colIt = colBegin; colIt != colEnd; ++colIt) {
		int col = colIt->first;
		double entry = colIt->second;
		elementMatrix[row][col] += entry;
	      }
	    }
	  }

	  // treat unique cols
	  std::set<DegreeOfFreedom> elementColDOFs;
	  rowsBegin = uniqueColMatrix_.begin();
	  rowsEnd = uniqueColMatrix_.end();
	  for(rowsIt = rowsBegin; rowsIt != rowsEnd; ++rowsIt) {
	    int row = rowsIt->first;
	    colBegin = rowsIt->second.begin();
	    colEnd = rowsIt->second.end();
	    for(colIt = colBegin; colIt != colEnd; ++colIt) {
	      int col = colIt->first;
	      DegreeOfFreedom colDOF = *(colDOFPtrs[col]);
	      if(visitedColDOFs[colDOF] == 0) {
		elementColDOFs.insert(colDOF);
		double entry = colIt->second;
		elementMatrix[row][col] += entry;
	      }
	    }
	  }

	  std::set<DegreeOfFreedom>::iterator setIt;
	  std::set<DegreeOfFreedom>::iterator setBegin = elementColDOFs.begin();
	  std::set<DegreeOfFreedom>::iterator setEnd = elementColDOFs.end();
	  for(setIt = setBegin; setIt != setEnd; ++setIt) {
	    visitedColDOFs[*setIt] = 1;
	  }

	  // treat other dofs
	  rowsBegin = matrix_.begin();
	  rowsEnd = matrix_.end();
	  for(rowsIt = rowsBegin; rowsIt != rowsEnd; ++rowsIt) {
	    int row = rowsIt->first;
	    colBegin = rowsIt->second.begin();
	    colEnd = rowsIt->second.end();
	    for(colIt = colBegin; colIt != colEnd; ++colIt) {
	      int col = colIt->first;
	      double entry = colIt->second;
	      elementMatrix[row][col] += entry;
	    }
	  }

	  dest->addElementMatrix(1.0, elementMatrix, NULL);
	} 	
      }
      elInfo = stack.traverseNext(elInfo);
    }

    // set matrix to identity for other dofs
    DOFMatrix::Iterator destIt(dest, USED_DOFS);
    for(destIt.reset(); !destIt.end(); ++destIt) {
      if(destIt->size() == 0) {
	MatEntry matEntry;
	matEntry.col = destIt.getDOFIndex();
	matEntry.entry = 1.0;
	destIt->push_back(matEntry);
      }
    }
  }

  void InterpolRestrictMatrix::print()
  {
    FUNCNAME("InterpolRestrictMatrix::print()");

    if(!initialized_) {
      initialized_ = true;
      fillMemory();
    }

    std::map<int, std::map<int, double> >::iterator rowsIt;
    std::map<int, std::map<int, double> >::iterator rowsBegin;
    std::map<int, std::map<int, double> >::iterator rowsEnd;
    std::map<int, double>::iterator colIt;
    std::map<int, double>::iterator colBegin;
    std::map<int, double>::iterator colEnd;
    std::vector<DegreeOfFreedom>::iterator dofIt;
    std::vector<DegreeOfFreedom>::iterator dofBegin = directCopyDOFs_.begin();
    std::vector<DegreeOfFreedom>::iterator dofEnd = directCopyDOFs_.end();


    // treat direct-copy dofs
    MSG("direct copy dofs:\n");
    for(dofIt = dofBegin; dofIt != dofEnd; ++dofIt) {
      MSG("%d\n", *dofIt);
    }

    // treat unique rows
    MSG("unique row matrix:\n");
    rowsBegin = uniqueRowMatrix_.begin();
    rowsEnd = uniqueRowMatrix_.end();
    for(rowsIt = rowsBegin; rowsIt != rowsEnd; ++rowsIt) {
      int row = rowsIt->first;
      colBegin = rowsIt->second.begin();
      colEnd = rowsIt->second.end();
      for(colIt = colBegin; colIt != colEnd; ++colIt) {
	int col = colIt->first;
	double entry = colIt->second;
	MSG("uniqueRow[%d][%d] = %e\n", row, col, entry);
      }
    }

    // treat unique cols
    MSG("unique col matrix:\n");
    rowsBegin = uniqueColMatrix_.begin();
    rowsEnd = uniqueColMatrix_.end();
    for(rowsIt = rowsBegin; rowsIt != rowsEnd; ++rowsIt) {
      int row = rowsIt->first;
      colBegin = rowsIt->second.begin();
      colEnd = rowsIt->second.end();
      for(colIt = colBegin; colIt != colEnd; ++colIt) {
	int col = colIt->first;
	double entry = colIt->second;
	MSG("uniqueCol[%d][%d] = %e\n", row, col, entry);
      }
    }

    // treat other dofs
    MSG("matrix:\n");
    rowsBegin = matrix_.begin();
    rowsEnd = matrix_.end();
    for(rowsIt = rowsBegin; rowsIt != rowsEnd; ++rowsIt) {
      int row = rowsIt->first;
      colBegin = rowsIt->second.begin();
      colEnd = rowsIt->second.end();
      for(colIt = colBegin; colIt != colEnd; ++colIt) {
	int col = colIt->first;
	double entry = colIt->second;
	MSG("matrix[%d][%d] = %e\n", row, col, entry);
      }
    }

    // print row dof ptrs
    MSG("row DOF ptrs:\n");
    int i;
    for(i = 0; i < numRows_; i++) {
      MSG("row %d : %p\n", i, rowDOFPtrs_[i]);
    }

    // print col dof ptrs
    MSG("col DOF ptrs:\n");
    for(i = 0; i < numCols_; i++) {
      MSG("col %d : %p\n", i, colDOFPtrs_[i]);
    }

    // print row dof ptrs
    if(type0_used_) {
      MSG("type0 row DOF ptrs:\n");
      int i;
      for(i = 0; i < numRows_; i++) {
	MSG("row %d : %p\n", i, type0_rowDOFPtrs_[i]);
      }

      // print col dof ptrs
      MSG("type0 col DOF ptrs:\n");
      for(i = 0; i < numCols_; i++) {
	MSG("col %d : %p\n", i, type0_colDOFPtrs_[i]);
      }
    }
  }
}
