// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  crystal growth group                                                  ==
// ==                                                                        ==
// ==  Stiftung caesar                                                       ==
// ==  Ludwig-Erhard-Allee 2                                                 ==
// ==  53175 Bonn                                                            ==
// ==  germany                                                               ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  http://www.caesar.de/cg/AMDiS                                         ==
// ==                                                                        ==
// ============================================================================

/** \file NonLinUpdater.h */

#ifndef AMDIS_NONLINAPDATOR_H
#define AMDIS_NONLINAPDATOR_H

#include "MemoryManager.h"
#include "SystemVector.h"
#include "MatrixVector.h"

namespace AMDiS {

  class DOFMatrix;
  template<typename T> class DOFVector;

  // =========================================================================
  // ===== class NonLinUpdater ===============================================
  // =========================================================================

  /** \brief
   * Used in non linear solvers for matrix and vector assemblage.
   * Base class for NonLinUpdaterScal and NonLinUpdaterVec.
   */
  template<typename Vector>
  class NonLinUpdater
  {
  public:
    virtual ~NonLinUpdater() {};

    /** \brief
     * Pure virtual update function.
     */
    virtual void update(bool updateDF,
			Vector *F) = 0;
  };

  // =========================================================================
  // ===== class NonLinUpdaterScal ===========================================
  // =========================================================================

  /** \brief
   * NonLinUpdater for scalar problems.
   */
  class NonLinUpdaterScal : public NonLinUpdater<DOFVector<double> >
  {
  public:
    MEMORY_MANAGED(NonLinUpdaterScal);

    /** \brief
     * Constructor.
     */
    NonLinUpdaterScal(DOFMatrix *m) : df(m) {};

    virtual ~NonLinUpdaterScal() {};

    /** \brief
     * Implementation of NonLinUpdater::update().
     */
    virtual void update(bool updateDF,
			DOFVector<double> *F);

  protected:
    /** \brief
     * Matrix to be updated.
     */
    DOFMatrix *df;
  };

  // =========================================================================
  // ===== class NonLinUpdaterVec ============================================
  // =========================================================================

  /** \brief
   * NonLinUpdater for vector valued problems.
   */
  class NonLinUpdaterVec : public NonLinUpdater<SystemVector>
  {
  public:
    MEMORY_MANAGED(NonLinUpdaterVec);

    /** \brief
     * Constructor.
     */
    NonLinUpdaterVec(Matrix<DOFMatrix*> *m) : df(m) {};

    virtual ~NonLinUpdaterVec() {};

    /** \brief
     * Implementation of NonLinUpdater::update().
     */
    virtual void update(bool updateDF,
			SystemVector *F);

  protected:
    /** \brief
     * Matrix to be updated.
     */
    Matrix<DOFMatrix*> *df;
  };

}

#endif
